# Learning-Guided Rolling Horizon Optimization for Long-Horizon Flexible Job Shop Scheduling

This repository contains the code to our Neurips 2024 submission titled `Learning-Guided Rolling Horizon Optimization for Long-Horizon Flexible Job Shop Scheduling`. 

This README file provides instructions on environment setup, FJSP instance generation, RHO paramter search, training data collection, model training and testing.

## Environment Setup

Our implementation uses python 3.9.16, Pytorch 2.0.0, OR-Tools 9.9.3963. The other package dependencies are listed in `requirement.txt` and can be installed with the following command:

```
pip install -r requirements.txt
```

## FJSP Instance Generation
We can generate FJSP instances with [N_MACHINES] number of machines, [N_JOBS] number of jobs, [N_TASKS_PER_JOB] number of operations per job as follows. The FJSP objective is specified by OPTIM_OPION [start_delay | start_and_end_delay]. In our experiments, we generate 600 instances for each FJSP setting, with index 0-9 for RHO parameter search, 10-459 for training, 460-479 for validation, and 500-599 for testing.

```
export N_MACHINES=25
export N_JOBS=25
export N_TASKS_PER_JOB=25
export OPTIM_OPTION='start_delay'
export DATA_IDX_START=0
export DATA_IDX_END=600  # the number of instances to generate

python flexible_jss_instance.py --n_machines $N_MACHINES --n_jobs $N_JOBS \
        --n_tasks_per_job $N_TASKS_PER_JOB --optim_option $OPTIM_OPTION \
        --data_index_start $DATA_IDX_START --data_index_end $DATA_IDX_END
```

## RHO Parameter Search 
As detailed in Appendix A.4.3, we perform a grid search to find the best parameter (Planning Horizon Size $H$, Execution Step Size $S$, Subproblem Time Limit $T$, Subproblem Early Termination Time $T_{es}$) for each method (Default RHO, Oracle, First and Random baselines). The following script evaluates the performance on one combination of the parameters, specified by [PARAM_IDX] (39 parameters in total). The performance statistics (objective and solve time) are saved in the `$PARAM_STATS_DIR` directory.

```
export N_MACHINES=25
export N_JOBS=25
export N_TASKS_PER_JOB=25
export OPTIM_OPTION='start_delay'

# the index of the parameter to evaluate the performance; from 0-38
export PARAM_IDX=0  
# the indices of FJSP instances to perform parameter search
export DATA_IDX_START=0
export DATA_IDX_END=10  

export PARAM_DIR='train_test_dir/param_search/params/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct08-madapt-task-machine-cf3'
# The directory to save parameter statistics to
export PARAM_STATS_DIR='train_test_dir/param_search/stats/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct08-madapt-task-machine-cf3'
# FJSP instance directory
export FJSP_DATA_DIR='train_data_dir/instance/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3'

python flexible_jss_parameter_search.py --n_machines $N_MACHINES --n_jobs $N_JOBS \
        --n_tasks_per_job $N_TASKS_PER_JOB --optim_option $OPTIM_OPTION --param_idx $PARAM_IDX \
        --param_dir  $PARAM_DIR --param_stats_dir $PARAM_STATS_DIR --jss_data_dir $FJSP_DATA_DIR \
        --data_index_start $DATA_IDX_START --data_index_end $DATA_IDX_END
```

After which, we can find the best combination of RHO parameters for each method using the following line search procedure, where the best combination will be saved to `$PARAM_STATS_DIR/best_params.pkl`.

```
export N_MACHINES=25
export N_JOBS=25
export N_TASKS_PER_JOB=25
export OPTIM_OPTION='start_delay'

export PARAM_IDX=0
export DATA_IDX_START=0
export DATA_IDX_END=20

export PARAM_DIR='train_test_dir/param_search/params/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct08-madapt-task-machine-cf3'
export PARAM_STATS_DIR='train_test_dir/param_search/stats/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct08-madapt-task-machine-cf3'
export FJSP_DATA_DIR='train_data_dir/instance/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3'

python find_best_rho_param.py --n_machines $N_MACHINES --n_jobs $N_JOBS \
        --n_tasks_per_job $N_TASKS_PER_JOB --optim_option $OPTIM_OPTION \
        --param_dir  $PARAM_DIR --param_stats_dir $PARAM_STATS_DIR --jss_data_dir $FJSP_DATA_DIR \
        --data_index_start $DATA_IDX_START --data_index_end $DATA_IDX_END
```

## Training Data Collection (Look-ahead Oracle)
We can use the following script to generate training data from the assignment-based look-ahead oracle described in the main paper Sec. 4 (Fig. 2a). `$DATA_IDX` specifies the index of the FJSP instance that we want to generate the training data,  (`WINDOW`, `STEP`, `TIME_LIMIT`, `EARLY_STOP_TIME`) specifies Oracle's best RHO parameter combination for the FJSP setting, obtained from the previous parameter search procedure. The training data is saved to `$TRAIN_DATA_DIR`.


```
# index of the FJSP instance to generate the training data (input and oracle labels)
export DATA_IDX=10 

# FJSP setting 
export N_MACHINES=25
export N_JOBS=25
export N_TASKS_PER_JOB=25
export OPTIM_OPTION='start_delay'

# Oracle's best parameter combination for the FJSP setting
export WINDOW=80
export STEP=25
export TIME_LIMIT=60
export EARLY_STOP_TIME=3

# directory to save the training data
export TRAIN_DATA_DIR='train_data_dir/train_data/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3'
# FJSP instance directory
export FJSP_DATA_DIR='train_data_dir/instance/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3'
# directory to save data collection statistics
export stats_dir='train_data_dir/stats/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3'


python flexible_jss_main.py --data_idx $DATA_IDX  \
        --n_machines $N_MACHINES --n_jobs $N_JOBS \
        --n_tasks_per_job $N_TASKS_PER_JOB --optim_option $OPTIM_OPTION \
        --window $WINDOW --step $STEP --time_limit $TIME_LIMIT --stop_search_time $EARLY_STOP_TIME \
        --train_data_dir $TRAIN_DATA_DIR  --jss_data_dir $FJSP_DATA_DIR \
        --stats_dir $stats_dir --load_data
```

## Model Training
We can use the following script to train the model, which will be saved to `$MODEL_DIR`. We need to specify the FJSP instance directory in `$TRAIN_DATA_DIR` and the training data directory in `$TRAIN_DATA_DIR`. We can further specify the training and validation split through the indices `$DATA_INDEX_START`, `$DATA_INDEX_END`, `$VAL_INDEX_START`, `$VAL_INDEX_END`, and the number of training epochs through `$EPOCHS`.

```
# training data indices
export DATA_INDEX_START=10
export DATA_INDEX_END=460
# validation data indices
export VAL_INDEX_START=460
export VAL_INDEX_END=480
# number of training epochs
export EPOCHS=600
# scaling factor for the positive labels (fixed operations)
export POS_WEIGHT=0.5
export MODEL_NAME=model_pw0.5

# FJSP setting 
export N_MACHINES=25
export N_JOBS=25
export N_TASKS_PER_JOB=30
export OPTIM_OPTION='start_delay'

# Oracle's best parameter combination for the FJSP setting
export WINDOW=80
export STEP=25
export TIME_LIMIT=60
export EARLY_STOP_TIME=3

# directory that contains the training data 
export TRAIN_DATA_DIR=train_data_dir/train_data/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3
# directory that contains the FJSP instances
export FJSP_DATA_DIR=train_data_dir/instance/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3
export LOG_DIR=train_test_dir/train_log/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3
export TEST_STATS_DIR=train_test_dir/test_stats/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3
# directory to save the model
export MODEL_DIR=train_test_dir/model/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3


python flexible_jss_learn.py \
        --data_index_start $DATA_INDEX_START --data_index_end $DATA_INDEX_END \
        --val_index_start $VAL_INDEX_START --val_index_end $VAL_INDEX_END --num_epochs $EPOCHS --model_name $MODEL_NAME --pos_weight $POS_WEIGHT \
        --train_data_dir $TRAIN_DATA_DIR --jss_data_dir $FJSP_DATA_DIR \
        --log_dir $LOG_DIR --test_stats_dir $TEST_STATS_DIR --model_dir $MODEL_DIR \
        --n_machines $N_MACHINES --n_jobs $N_JOBS \
        --n_tasks_per_job $N_TASKS_PER_JOB --optim_option $OPTIM_OPTION \
        --window $WINDOW --step $STEP --time_limit $TIME_LIMIT --stop_search_time $EARLY_STOP_TIME
```

## Testing
Finally, we can use the following script to evaluate the test performances for Default RHO, First and Random RHO baselines, our learning method L-RHO (following the inference procedure described in the main paper Sec. 4, Fig 2b), and solving the full FJSP without decomposition. We need to specify the test data indices through `$TEST_INDEX_START` and `$TEST_INDEX_END`, the FJSP instance directory `$FJSP_DATA_DIR`, the pretrained model directory `$MODEL_DIR`, and the path to the best RHO parameter file `$BEST_PARAM_FILE`. The test performance statistics (objective and solve time) will be saved to `$TEST_STATS_DIR`.

```
# test data indices
export TEST_INDEX_START=500
export TEST_INDEX_END=600

# training data indices (not used)
export DATA_INDEX_START=0
export DATA_INDEX_END=450

# pretrained model name
export MODEL_NAME=model_pw0.5
export POS_WEIGHT=0.5

# which epoch to load the pretrained model
export LOAD_MODEL_EPOCH=120

# FJSP setting 
export N_MACHINES=25
export N_JOBS=25
export N_TASKS_PER_JOB=30
export OPTIM_OPTION='start_delay'

# Oracle's best parameter combination for the FJSP setting 
# (only used if BEST_PARAM_FILE below does not exist)
export WINDOW=80
export STEP=25
export TIME_LIMIT=60
export EARLY_STOP_TIME=3

# directory that contains the training data 
export TRAIN_DATA_DIR=train_data_dir/train_data/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3
# directory that contains the FJSP instances
export FJSP_DATA_DIR=train_data_dir/instance/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3
export LOG_DIR=train_test_dir/train_log/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3

# directory to save the test performance statistics 
export TEST_STATS_DIR=train_test_dir/test_stats/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3
# directory that contains the pretrained model
export MODEL_DIR=train_test_dir/model/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3-w80-s25-t60-st3
# path to the best RHO parameter file (obtained from the above parameter search procedure)
export BEST_PARAM_FILE='train_test_dir/param_search/params/m25-j25-t25-p1-lo5-hi20-s0-h15-p-ct0.8-madapt-task-machine-cf3/best_params.pkl'

python flexible_jss_learn.py --data_index_start $DATA_INDEX_START --data_index_end $DATA_INDEX_END \
        --val_index_start $TEST_INDEX_START --val_index_end $TEST_INDEX_END --model_name $MODEL_NAME \
        --pos_weight $POS_WEIGHT --test --load_model_epoch $LOAD_MODEL_EPOCH  \
        --n_machines $N_MACHINES --n_jobs $N_JOBS \
        --n_tasks_per_job $N_TASKS_PER_JOB --optim_option $OPTIM_OPTION \
        --window $WINDOW --step $STEP --time_limit $TIME_LIMIT --stop_search_time $EARLY_STOP_TIME \
        --train_data_dir $TRAIN_DATA_DIR --jss_data_dir $FJSP_DATA_DIR \
        --log_dir $LOG_DIR --test_stats_dir $TEST_STATS_DIR --model_dir $MODEL_DIR \
        --load_best_params --best_params_file $BEST_PARAM_FILE
```
